use std::str::FromStr;

use palette::{
    cam16::{Cam16Jch, Parameters},
    IntoColor, Srgb,
};

use crate::errors;

#[derive(Debug, Clone)]
pub struct TonalPalette {
    pub key_color: Cam16Jch<f32>,
}

#[inline]
fn approximately_equal(a: f32, b: f32) -> bool {
    const EPSILON: f32 = 0.000001;
    (a - b).abs() < EPSILON
}

#[inline]
fn find_max_chroma(cache: &mut Vec<(f32, f32)>, hue: f32, tone: f32) -> f32 {
    for (k, v) in cache.iter() {
        if approximately_equal(*k, tone) {
            return *v;
        }
    }
    let chroma = Cam16Jch::new(tone, 200.0, hue).chroma;
    cache.push((tone, chroma));
    chroma
}

fn from_hue_and_chroma(hue: f32, chroma: f32) -> Cam16Jch<f32> {
    let mut max_chroma_cache = Vec::new();
    let hue = if hue >= 360.0 { hue - 360.0 } else { hue };
    const PIVOT_TONE: f32 = 50.0;
    const TONE_STEP_SIZE: f32 = 1.0;
    const EPSILON: f32 = 0.01;

    let mut lower_tone = 0.0_f32;
    let mut upper_tone = 100.0_f32;
    while lower_tone < upper_tone {
        let mid_tone = ((lower_tone + upper_tone) / 2.0).floor();
        let is_ascending = find_max_chroma(&mut max_chroma_cache, hue, mid_tone)
            < find_max_chroma(&mut max_chroma_cache, hue, mid_tone + TONE_STEP_SIZE);
        let sufficient_chroma =
            find_max_chroma(&mut max_chroma_cache, hue, mid_tone) >= chroma - EPSILON;

        if sufficient_chroma {
            if (lower_tone - PIVOT_TONE).abs() < (upper_tone - PIVOT_TONE).abs() {
                upper_tone = mid_tone;
            } else {
                if approximately_equal(lower_tone, mid_tone) {
                    return Cam16Jch::new(lower_tone, chroma, hue);
                }
                lower_tone = mid_tone;
            }
        } else {
            if is_ascending {
                lower_tone = mid_tone + TONE_STEP_SIZE;
            } else {
                upper_tone = mid_tone;
            }
        }
    }
    Cam16Jch::new(lower_tone, chroma, hue)
}

impl TryFrom<String> for TonalPalette {
    type Error = errors::ColorError;

    fn try_from(value: String) -> Result<Self, Self::Error> {
        let key_color = Cam16Jch::from_xyz(
            Srgb::from_str(&value)
                .map_err(|_| errors::ColorError::UnrecogniazedRGB(value))?
                .into_format::<f32>()
                .into_color(),
            Parameters::default_static_wp(40.0),
        );
        Ok(TonalPalette { key_color })
    }
}

impl TonalPalette {
    pub fn from_hue_and_chroma(hue: f32, chroma: f32) -> Self {
        let key_color = from_hue_and_chroma(hue, chroma);
        TonalPalette { key_color }
    }

    pub fn tone(&self, tone: f32) -> Cam16Jch<f32> {
        let toned_color = Cam16Jch::new(tone, self.key_color.chroma, self.key_color.hue);

        toned_color
    }
}
